# ---------------------------------------------------------------------
# Copyright (c) 2017 Kim Peter Wabersich, email: wabersich@kimpeter.de
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
# -----------------------------------------------------------------
import numpy as np
import matplotlib.pyplot as plt
import scipy.io as sio
import os


def empc_closed_loop_simulation(ges_problem_robust, simulation, plotter, graph_robust_empc_solver,
                                real_disturbance=False, save_path = ''):

    # init real system trajectory with initial state
    x_real_system = ges_problem_robust.x0_real_.reshape([ges_problem_robust.num_state_variables_real_, 1])
    x_graph_system = ges_problem_robust.x0_graph_
    u_real_system = np.array([[]])
    u_graph_system = np.array([])

    ## Closed loop simulation
    for n in range(0, 5):
        # solve problem
        optimal_trajectory = graph_robust_empc_solver.solve_empc(verbose=False)

        # apply k-inputs (periodic operation is optimal)
        for nn in range(0, ges_problem_robust.P_):
            simulation.simulate(optimal_trajectory.u_real_[:, nn], optimal_trajectory.u_graph_[nn],
                                robust=real_disturbance)

        # print avg cost
        simulation.calculate_avg_cost(False)

        # set starting condition
        ges_problem_robust.x0_real_ = simulation.x_real_t_
        ges_problem_robust.x0_graph_ = simulation.x_graph_t_
        graph_robust_empc_solver.set_inp(ges_problem_robust)


    ## visualize problem
    # plot optimization solution
    plotter.plot(optimal_trajectory.x_real_, optimal_trajectory.x_graph_,
                                        optimal_trajectory.u_real_, optimal_trajectory.u_graph_,
                                        ges_problem_robust.planning_horizon_n_, "optimization")

    # plot real system trajectory
    plotter.plot(simulation.x_real_history_, simulation.x_graph_history_,
                                        simulation.u_real_history_, simulation.u_graph_history_,
                                        simulation.u_real_history_.shape[1], "real",
                                        close_plot_first = False)

    if not save_path == '':
        # save trajectory
        print('save results...')
        if not os.path.exists(save_path):
            os.makedirs(save_path)
        sio.savemat(save_path+"x_real.mat", {'x_real_': simulation.x_real_history_})
        sio.savemat(save_path+"x_graph.mat", {'x_graph_': simulation.x_graph_history_})
        sio.savemat(save_path+"u_real.mat", {'u_real_': simulation.u_real_history_})
        sio.savemat(save_path+"u_graph.mat", {'u_graph_': simulation.u_graph_history_})

    plt.show(block=True)
