# ---------------------------------------------------------------------
# Copyright (c) 2017 Kim Peter Wabersich, email: wabersich@kimpeter.de
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
# -----------------------------------------------------------------
import numpy as np
import copy
from integer_grid.IntegerGrid import IntegerGrid
from graph_robust_empc_solver import InterfaceRobustGes
import mpctools as mpc
from casadi import *


# In the following the nominal states and inputs are x and u in order to
# be consistent with mpc-tools. Real, disturbed states and corresponding inputs are x_dist, u_dist.
class GraphRobustEmpcSolver(object):
    def __init__(self, ges_problem_robust, robust=True):
        # robust empc mode
        self.robust_ = robust
        # current input graph trajectory for which we solve the problem
        self.u_graph_trajectory_ = 0
        # current state graph trajectory for which we solve the problem
        self.x_graph_trajectory_ = 0

        # casadi function of model dynamics
        self.f_casadi = 0

        # casadi function of integrated stage cost
        self.stage_cost_casadi = 0

        # casadi function of inequality constraints
        self.ineq_constraints_casadi = 0

        # terminal cost function
        self.V_f_casadi = 0

        # terminal set constraint
        self.X_f_con_casadi = 0

        # mpc-tools controller
        self.controller_ = 0

        # initial guess
        self.initial_guess_ = 0

        # set model and initialize
        self.set_inp(ges_problem_robust)


    # set new problem specification
    def set_inp(self, interface_ges):
        self.inp_ = interface_ges
        self.init_graph_dynamics()
        self.init()

    # main inits
    def init(self):
        self.propagate_state_graph_dynamics()
        self.init_mpc_tools()
        self.make_initial_guess()
        self.init_solver()

    # initialize graph dynamics
    def init_graph_dynamics(self):
        number_of_u_graph_values = self.inp_.num_graph_input_values_
        lower_bound_u_graph = [0] * self.inp_.planning_horizon_n_
        upper_bound_u_graph = [number_of_u_graph_values - 1] * self.inp_.planning_horizon_n_
        self.u_graph_trajectory_ = IntegerGrid(self.inp_.planning_horizon_n_,
                                               lower_bound_u_graph,
                                               upper_bound_u_graph
                                               )
        self.x_graph_trajectory_ = np.zeros([self.inp_.planning_horizon_n_ + 1, 1])

    # Given a fixed graph trajectory this function allows for looking up the graph states w.r.t time
    def graph_look_up_table(self, t):
        graph_x_value = 0
        graph_u_value = 0
        for n in range(0, self.inp_.planning_horizon_n_):
            graph_x_value += if_else(n==t[0], self.x_graph_trajectory_[n],0)
            graph_u_value += if_else(n==t[0], self.u_graph_trajectory_.x_[n],0)
        graph_x_value += if_else(self.inp_.planning_horizon_n_ == t[0], self.x_graph_trajectory_[self.inp_.planning_horizon_n_], 0)
        return graph_x_value, graph_u_value

    # propagate graph dynamics and create cascadi look up function
    def propagate_state_graph_dynamics(self):
        for n in range(0, self.inp_.planning_horizon_n_):  # for every time stamp
            # propagate graph
            u_graph_n = self.u_graph_trajectory_.x_[n]
            self.x_graph_trajectory_[n + 1] = self.inp_.function_graph_dynamics_(self.x_graph_trajectory_[n, 0], u_graph_n)

    # brute force graph optimization step
    def brute_force_graph_step(self):
        # count integer grid (search tree)
        x_, end_reached = self.u_graph_trajectory_.count()
        return end_reached

    # decompose variables
    def decompose_states_inputs(self, x_real_t, u_real_x_slack):
        # extract state
        x_real = np.array(x_real_t[:self.inp_.num_state_variables_real_])
        # extract time
        t = x_real_t[self.inp_.num_state_variables_real_]
        # extract input
        u_real = np.array(u_real_x_slack[:self.inp_.num_input_variables_real_])
        # extract slack variable
        x_slack = u_real_x_slack[self.inp_.num_input_variables_real_]
        # look up graph state
        x_graph,u_graph = self.graph_look_up_table(t[0])

        return x_real, u_real, x_graph, u_graph, x_slack, t

    # differential equation for mpc-tools
    def f(self, x_real_t, u_real_x_slack):
        # decompose casadi variables
        x_real, u_real, x_graph, u_graph, x_slack, t = self.decompose_states_inputs(x_real_t, u_real_x_slack)
        # simulation step
        x_real_p = mpc.mtimes(self.inp_.function_matrix_a_dynamics_(x_graph),
                              x_real) + \
                   mpc.mtimes(self.inp_.function_matrix_b_dynamics_(x_graph), u_real) + \
                   self.inp_.function_disturbance_nominal_()
        return [x_real_p, t + 1]

    # stage cost is defined via slack variable
    def stage_cost(self, x_real_t, u_real_x_slack):
        # decompose casadi variables
        x_real, u_real, x_graph, u_graph, x_slack, t = self.decompose_states_inputs(x_real_t, u_real_x_slack)
        return x_slack

    # stage cost slack constraints
    def stage_cost_con(self,x_real_t, u_real_x_slack):
        x_real, u_real, x_graph, u_graph, x_slack, t = self.decompose_states_inputs(x_real_t, u_real_x_slack)
        x_slack_arr = np.ones(self.inp_.num_stage_cost_funcs_)*x_slack
        return self.inp_.function_arr_stage_cost_(u_real, x_real, u_graph, x_graph) - x_slack_arr

    # int stage cost slack constraints
    def int_stage_cost_con(self, x_real_t, u_real_x_slack):
        x_real, u_real, x_graph, u_graph, x_slack, t = self.decompose_states_inputs(x_real_t, u_real_x_slack)
        x_slack_arr = np.ones(self.inp_.num_int_stage_cost_funcs_) * x_slack
        return self.inp_.function_arr_int_stage_cost_arr_(u_real, x_real, u_graph, x_graph, self.inp_.RPI_set_lb_,
                                                          self.inp_.RPI_set_ub_) - x_slack_arr

    # terminal cost is also defined via slack variable
    def V_f(self, x_real_t):
        x_real, u_real, x_graph, u_graph, x_slack, t = self.decompose_states_inputs(
            x_real_t, np.zeros(self.inp_.num_input_variables_real_ + 1))
        # call terminal cost using the final extended state
        return self.inp_.function_V_f_(x_real, x_graph)  # plus cost

    def X_f_con(self, x_real_t):
        # decompose casadi variables
        x_real, u_real, x_graph, u_graph, x_slack, t = self.decompose_states_inputs(
            x_real_t, np.zeros(self.inp_.num_input_variables_real_ + 1))
        return self.inp_.function_arr_X_f_constraint_(x_real, x_graph)

    # init all cascadi functions via mpc-tools for nominal system solver
    def init_mpc_tools(self):
        # init dynamic system
        self.f_casadi = mpc.getCasadiFunc(self.f,
                                    [self.inp_.num_state_variables_real_ + 1 ,  # +time
                                     self.inp_.num_input_variables_real_ + 1],  # +slack
                                    ["x_real+t", "u_real"])

        # init slack based stage cost
        self.stage_cost_casadi = mpc.getCasadiFunc(self.stage_cost,
                                                   [self.inp_.num_state_variables_real_ + 1,  # +time
                                                    self.inp_.num_input_variables_real_ + 1],  # +slack
                                                   ["x_real+t", "u_real"])

        # terminal cost function
        self.V_f_casadi = mpc.getCasadiFunc(self.V_f,
                                            [self.inp_.num_state_variables_real_ + 1], # +time
                                            # +time
                                            ["x_real+t"])

        # terminal regino constraint
        self.X_f_con_casadi = mpc.getCasadiFunc(self.X_f_con,
                                                 [self.inp_.num_state_variables_real_ + 1],  # +time
                                                 ["x_real+t"], "ef")

        if self.robust_ == True:
            # slack variable stage cost constraint
            self.ineq_constraints_casadi = mpc.getCasadiFunc(self.int_stage_cost_con,
                                                             [self.inp_.num_state_variables_real_ + 1,  # +time
                                                              self.inp_.num_input_variables_real_ + 1],  # +slack
                                                             ["x_real+t", "u_real"], "e")



        else:
            # slack variable stage cost constraint
            self.ineq_constraints_casadi = mpc.getCasadiFunc(self.stage_cost_con,
                                      [self.inp_.num_state_variables_real_ + 1,  # +time
                                      self.inp_.num_input_variables_real_ + 1],  # +slack
                                      ["x_real+t", "u_real"],"e")


    def init_solver(self):
        # init bounds
        if self.robust_ == True:
            lb = {
                "u": np.tile(np.concatenate((self.inp_.v_lb_tight_, np.array([-np.inf]))),  # input constraints plus slack
                             (self.inp_.planning_horizon_n_, 1)),
                "x": np.tile(np.concatenate((self.inp_.z_lb_tight_, np.array([0]))),        # state constraints plus time
                             (self.inp_.planning_horizon_n_+ 1, 1)),
            }
            ub = {
                "u": np.tile(np.concatenate((self.inp_.v_ub_tight_, np.array([np.inf]))),   # input constraints plus slack
                             (self.inp_.planning_horizon_n_, 1)),
                "x": np.tile(np.concatenate((self.inp_.z_ub_tight_, np.array([np.inf]))),   # state constraints plus time
                             (self.inp_.planning_horizon_n_ + 1, 1)),
            }
        else:
            lb = {
                "u": np.tile(np.concatenate((self.inp_.u_lb_, np.array([-np.inf]))),
                             # input constraints plus slack
                             (self.inp_.planning_horizon_n_, 1)),
                "x": np.tile(np.concatenate((self.inp_.x_lb_, np.array([0]))),  # state constraints plus time
                             (self.inp_.planning_horizon_n_ + 1, 1)),
            }
            ub = {
                "u": np.tile(np.concatenate((self.inp_.u_ub_, np.array([np.inf]))),
                             # input constraints plus slack
                             (self.inp_.planning_horizon_n_, 1)),
                "x": np.tile(np.concatenate((self.inp_.x_ub_, np.array([np.inf]))),  # state constraints plus time
                             (self.inp_.planning_horizon_n_ + 1, 1)),
            }
        udiscrete = np.array([False, False, False])
        # init solver
        if self.robust_ == True:
            N = {"x": self.inp_.num_state_variables_real_ + 1, # plus time
                 "u": self.inp_.num_input_variables_real_ + 1, # plus slack
                 "e": self.inp_.num_int_stage_cost_funcs_, # number of constraints
                 "t": self.inp_.planning_horizon_n_,
                 "Period": self.inp_.P_}
        else:
            N = {"x": self.inp_.num_state_variables_real_ + 1,  # plus time
                 "u": self.inp_.num_input_variables_real_ + 1,  # plus slack
                 "e": self.inp_.num_stage_cost_funcs_,          # number of constraints
                 "t": self.inp_.planning_horizon_n_,
                 "Period": self.inp_.P_}  # in case we have no end - state cost
        funcargs = {"f": ["x", "u"], "e": ["x","u"], "l": ["x","u"], "ef": ["x"]}
        self.controller_ = mpc.nmpc(self.f_casadi, self.stage_cost_casadi, N, e=self.ineq_constraints_casadi,
                                    lb=lb, ub=ub, Pf=self.V_f_casadi, ef=self.X_f_con_casadi,
                                    funcargs=funcargs, casaditype="SX", udiscrete=udiscrete,
                                    verbosity=0, isQP=False, guess=self.initial_guess_)
        self.controller_.initialize(solveroptions=dict(max_iter=10000))
        self.set_initial_condition()

    # create guess based on optimal orbit
    def make_initial_guess(self):
        x_init = np.zeros((self.inp_.num_state_variables_real_+1, self.inp_.planning_horizon_n_+1))
        u_init = np.zeros((self.inp_.num_input_variables_real_+1, self.inp_.planning_horizon_n_))
        n_period = 0
        # tigh optimal periodic orbits together as initial guess
        for n in range(0, self.inp_.planning_horizon_n_):
            if n_period >= self.inp_.P_:
                n_period = 0
            # states
            x_init[0:self.inp_.num_state_variables_real_, n] = self.inp_.x_orbit_opt_[:, n_period]
            # time
            x_init[self.inp_.num_state_variables_real_, n] = n
            # inputs
            u_init[0:self.inp_.num_input_variables_real_, n] = self.inp_.u_orbit_opt_[:, n_period]
            # stage cost slack (make feasible)
            u_init[self.inp_.num_input_variables_real_, n] = 1000.
            n_period += 1
        # end state
        x_init[0:self.inp_.num_state_variables_real_, self.inp_.planning_horizon_n_] = self.inp_.x_orbit_opt_[:, 0]
        # end time
        x_init[self.inp_.num_state_variables_real_, self.inp_.planning_horizon_n_] = self.inp_.planning_horizon_n_
        # set initial guess
        self.initial_guess_ = dict(x=x_init.transpose(), u=u_init.transpose())

    # set initial condition
    def set_initial_condition(self):
        self.controller_.fixvar("x", 0, np.concatenate((self.inp_.x0_real_, np.array([0]))))
        self.x_graph_trajectory_[0] = self.inp_.x0_graph_

    # main solver
    def solve_empc(self, verbose=False):
        brute_force_search = True  # True as long as any new posibility exists
        optimization_process_information_sol = []  # store solution variables
        optimization_process_information_x_graph = []  # store solution of x_graph variable
        optimization_process_information_u_graph = []  # store solution of u_graph variable
        optimization_process_information_objectives = []  # store objectives
        k=0
        while brute_force_search == 1:
            b_graph_constraints_satisfied = False
            if self.robust_ == True:
                # check tightened graph constraints
                b_graph_constraints_satisfied = self.inp_.function_check_tilde_graph_constraints_tight_(
                    self.x_graph_trajectory_, self.u_graph_trajectory_.x_)
            else:
                b_graph_constraints_satisfied = self.inp_.function_check_graph_constraints(self.x_graph_trajectory_,
                                                                                           self.u_graph_trajectory_.x_)

            # check graph constraints
            if b_graph_constraints_satisfied == True:
                # init solver with fixed graph trajectory
                self.init()
                # solve!
                self.controller_.solve()
                # check solution and store objective
                if not (self.controller_.stats["status"] == "Solve_Succeeded"): #or self.controller_.stats["status"] == "Solved_To_Acceptable_Level"):
                    warnings.warn("Problems while solving a specific graph configuration:")
                    warnings.warn(self.controller_.stats["status"])
                    optimization_process_information_objectives.append(np.inf)
                else:
                    optimization_process_information_objectives.append(self.controller_.obj)
                # store solution
                optimization_process_information_sol.append(copy.copy(self.controller_.vardict))
                optimization_process_information_x_graph.append(np.array(self.x_graph_trajectory_))
                optimization_process_information_u_graph.append(np.array(self.u_graph_trajectory_.x_))

            # iterate graph
            x_, end_reached = self.u_graph_trajectory_.count()
            self.propagate_state_graph_dynamics()
            if end_reached == 1:
                brute_force_search = 0
            # information
            if (verbose == True):
                print('Solve input graph configuration No.'+str(k)+':')
                print(self.u_graph_trajectory_.x_)
            k = k+1
        # find best solution
        idx_opt = np.argmin(optimization_process_information_objectives)
        # decompose solution
        optimal_trajectory = InterfaceRobustGes.InterfaceRobustGesRes()
        optimal_trajectory.objective_ = optimization_process_information_objectives[idx_opt]
        optimal_trajectory.x_graph_ = optimization_process_information_x_graph[idx_opt].flatten()
        optimal_trajectory.u_graph_ = optimization_process_information_u_graph[idx_opt]
        optimal_trajectory.x_real_ = optimization_process_information_sol[idx_opt]['x'][0:self.inp_.planning_horizon_n_+1,0:self.inp_.num_state_variables_real_].transpose()
        optimal_trajectory.u_real_ = optimization_process_information_sol[idx_opt]['u'][0:self.inp_.planning_horizon_n_,0:self.inp_.num_input_variables_real_].transpose()
        # optimal_trajectory.objective_aux_ = L_aux(optimal_trajectory.x_real_[:,0:self.inp_.P_],
        #                                           optimal_trajectory.x_real_[:,self.inp_.P_:2*self.inp_.P_],
        #                                           optimal_trajectory.u_real_[:,0:self.inp_.P_],
        #                                           self.inp_)
        print('OPTIMAL SOLUTION:')
        print('objective='+str(optimal_trajectory.objective_))
        # print('auxiliary stage cost='+str(optimal_trajectory.objective_aux_))
        if(verbose == True):
            print('x_graph_='+str(optimal_trajectory.x_graph_))
            print('u_graph_='+str(optimal_trajectory.u_graph_))
            print('x_real_=\n' + str(optimal_trajectory.x_real_))
            print('u_real_=\n' + str(optimal_trajectory.u_real_))
        if optimal_trajectory.objective_== np.inf:
            print('error during optimization!')
            print('Debug!')
            assert(0)
        # if optimal_trajectory.objective_<0:
        #     print('x_graph_=' + str(optimal_trajectory.x_graph_))
        #     print('u_graph_=' + str(optimal_trajectory.u_graph_))
        #     print('x_real_=\n' + str(optimal_trajectory.x_real_.transpose()))
        #     print('u_real_=\n' + str(optimal_trajectory.u_real_.transpose()))
        #     print('Debug!')
        #     input()

        # detect uniqueness approximately
        sorted_objectives = np.sort(optimization_process_information_objectives, axis=None)
        if len(sorted_objectives) > 1:
            if np.isclose(sorted_objectives[0], sorted_objectives[1], 0.00001, 0.00001):
                optimal_trajectory.unique_ = False
            else:
                optimal_trajectory.unique_ = True
        else:
            optimal_trajectory.unique_ = True

        # return result
        return optimal_trajectory
