# ---------------------------------------------------------------------
# Copyright (c) 2017 Kim Peter Wabersich, email: wabersich@kimpeter.de
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
# -----------------------------------------------------------------
import graph_robust_empc_solver.InterfaceRobustGes as InterfaceRobustGes
import supply_chain_model.configurations.without_terminal_constraints.robust_configurations as robust_configurations
from casadi import *


# define the model dynamics, depending on the graph variable
def function_matrix_a_dynamics(x_graph):
    A = np.array([
        [0, 0, 0, 0, 0, 0],
        [1, 1, 0, 0, 0, 0],
        [0, 0, 1, 0, 0, 0],
        [0, 0, 0, 1, 0, 0],
        [0, 0, 0, 0, 1, 0],
        [0, 0, 0, 0, 0, 1],
    ])
    return A


# define input matrix
def function_matrix_b_dynamics(x_graph):
    at_supplier = if_else(x_graph==0, 1, 0)
    at_retailer1 = if_else(x_graph==1, 1, 0)
    at_retailer2 = if_else(x_graph==2, 1, 0)
    at_retailer3 = if_else(x_graph==3, 1, 0)
    B = np.array([
        [0, 1],
        [-at_supplier, 0],
        [1., 0],
        [-at_retailer1, 0],
        [-at_retailer2, 0],
        [-at_retailer3, 0]
    ])
    return B


# define graph dynamics
def function_graph_dynamics(x_graph, u_graph):
    # truck at supplier
    if x_graph == 0 and u_graph == 0:
        return 0
    if x_graph == 0 and u_graph == 1:
        return 1
    if x_graph == 0 and u_graph == 2:
        return 2
    # truck retailer 1
    if x_graph == 1:
        return 2
    # truck at retailer 2
    if x_graph == 2 and u_graph == 2:
        return 0
    elif x_graph == 2:
        return 3
    # truck at retailer 3
    if x_graph == 3:
        return 0


# define disturbances
def function_disturbance_nominal():
    return np.array([
            0,
            0,
            0,
            -1.,
            -1.,
            -1.])

# disturbance
def function_disturbance():
    return np.array([
            0,
            0,
            0,
            np.random.randint(-1, 1),
            np.random.randint(-1, 1),
            np.random.randint(-1, 1)])


def check_graph_constraints_dummy(x_graph_trajectory, u_graph_trajectory):
    return True

def check_graph_constraints(x_graph_trajectory, u_graph_trajectory):
    for n in range(0, len(u_graph_trajectory)):
        # remove redundant input cases
        if x_graph_trajectory[n] == 1 and u_graph_trajectory[n] == 1:
            return 0
        if x_graph_trajectory[n] == 1 and u_graph_trajectory[n] == 2:
            return 0
        if x_graph_trajectory[n] == 2 and u_graph_trajectory[n] == 1:
            return 0
        if x_graph_trajectory[n] == 3 and u_graph_trajectory[n] == 1:
            return 0
        if x_graph_trajectory[n] == 1 and u_graph_trajectory[n] == 2:
            return 0
    return 1

def function_check_tilde_graph_constraints_tight(z_graph_trajectory, v_graph_trajectory):
    if np.max(v_graph_trajectory) == 1 and np.min(v_graph_trajectory) == 1:
        return True
    else:
        return False

def function_check_tilde_graph_constraints_tight(z_graph_trajectory, v_graph_trajectory):
    optimal_graph_trajectory = np.array([0, 1])
    n_period_count = 0
    for z_graph in z_graph_trajectory:
        if n_period_count > 3:
            n_period_count = 0
        if not(z_graph == optimal_graph_trajectory[n_period_count]):
            return False
        n_period_count += 1
    return True


def check_graph_periodic_orbit(x_graph_trajectory, u_graph_trajectory):
    if x_graph_trajectory[0] == x_graph_trajectory[x_graph_trajectory.shape[0]-1]:
        return True
    else:
        return False


def function_arr_stage_cost_non_casadi(v_real, z_real, v_graph, z_graph):
    constant_term = z_real[0]       # supplier production cost
    constant_term+= 0.5*z_real[1]   # supplier storage cost
    constant_term+= z_real[2]       # truck storage cost
    # truck is driving cost
    if v_graph == 1:
        constant_term += 1
    if v_graph == 2:
        constant_term += 2
    # retailer cost
    return np.array([   z_real[3]+   z_real[4]+   z_real[5]+constant_term,
                        z_real[3]+   z_real[4]-10*z_real[5]+constant_term,
                        z_real[3]-10*z_real[4]+   z_real[5]+constant_term,
                        z_real[3]-10*z_real[4]-10*z_real[5]+constant_term,
                    -10*z_real[3]+   z_real[4]+   z_real[5]+constant_term,
                    -10*z_real[3]+   z_real[4]-10*z_real[5]+constant_term,
                    -10*z_real[3]-10*z_real[4]+   z_real[5]+constant_term,
                    -10*z_real[3]-10*z_real[4]-10*z_real[5]+constant_term
                        ])


def function_arr_stage_cost_(v_real, z_real, v_graph, z_graph):
    constant_term = z_real[0]       # supplier production cost
    constant_term+= 0.5*z_real[1]   # supplier storage cost
    constant_term+= z_real[2]       # truck storage cost
    # truck is driving cost
    constant_term += if_else(v_graph == 1, 1, 0)
    constant_term += if_else(v_graph == 2, 1, 0)
    # retailer cost
    return np.array([   z_real[3]+   z_real[4]+   z_real[5]+constant_term,
                        z_real[3]+   z_real[4]-10*z_real[5]+constant_term,
                        z_real[3]-10*z_real[4]+   z_real[5]+constant_term,
                        z_real[3]-10*z_real[4]-10*z_real[5]+constant_term,
                    -10*z_real[3]+   z_real[4]+   z_real[5]+constant_term,
                    -10*z_real[3]+   z_real[4]-10*z_real[5]+constant_term,
                    -10*z_real[3]-10*z_real[4]+   z_real[5]+constant_term,
                    -10*z_real[3]-10*z_real[4]-10*z_real[5]+constant_term
                        ])

def function_V_f_dummy(x_real, x_graph):
    return SX(0)

def function_arr_X_f_constraint_dummy(x_real, x_graph):
    return SX(0)


## definition of the robust model description
def get_supply_chain_robust_model(robust='none'):

    # interface for robust empc problem specification
    ges_problem = InterfaceRobustGes.InterfaceRobustGesReq()

    ## nominal configuration
    # dimensions
    ges_problem.num_input_variables_real_ = 2
    ges_problem.num_state_variables_real_ = 6
    ges_problem.num_graph_input_values_ = 3

    # state box constraints
    ges_problem.x_lb_ = np.array([0.0,      # supplier items in production >=0
                                  0.0,      # supplier storage >=0
                                  0.0,      # truck storage >=0
                                  -np.inf,
                                  -np.inf,
                                  -np.inf]) # retailer in RR
    ges_problem.x_ub_ = np.array([np.inf,   # supplier items in production >=0
                                  np.inf,   # supplier storage >=0
                                  20,       # truck storage <=10
                                  np.inf,
                                  np.inf,
                                  np.inf])  # retailer in RR

    # input box constraints
    ges_problem.u_lb_ = np.array([-np.inf,  # truck loading bounded by state bounds
                                  0])       # supplier production >=0

    ges_problem.u_ub_ = np.array([np.inf,   # supplier production >=0
                                  np.inf])  # truck loading bounded by state bounds

    # graph dynamics constraints
    ges_problem.function_check_graph_constraints = check_graph_constraints

    # graph dynamics
    ges_problem.function_graph_dynamics_ = function_graph_dynamics

    # graph dependent linear dynamics
    ges_problem.function_matrix_a_dynamics_ = function_matrix_a_dynamics
    ges_problem.function_matrix_b_dynamics_ = function_matrix_b_dynamics

    ges_problem.function_disturbance_nominal_ = function_disturbance_nominal
    ges_problem.function_disturbance_ = function_disturbance

    # stage cost
    ges_problem.num_stage_cost_funcs_ = 8
    ges_problem.function_arr_stage_cost_ = function_arr_stage_cost_
    ges_problem.function_arr_stage_cost_non_casadi_ = function_arr_stage_cost_non_casadi

    # terminal cost
    ges_problem.function_V_f_ = function_V_f_dummy

    # terminal region
    ges_problem.function_arr_X_f_constraint_ = function_arr_X_f_constraint_dummy

    # define initial condition
    # ges_problem.x0_real_ = np.array([0.0, 0.0, 0.0, -22.0, 20.0, -10.0])
    ges_problem.x0_real_ = np.array([0.0, 12.0, 20.0, 3.0, 3.0, 3.0])
    ges_problem.x0_graph_ = np.array(0)

    # define optimal periodic orbit
    ges_problem.x_orbit_opt_ = np.array([[12.0, 0.0, 0.0, 1.0, 2.0, 3.0],
                                         [0, 0, 12, 0, 1, 2],
                                         [0, 0, 8, 3, 0, 1],
                                         [0, 0, 4, 2, 3, 0]])
    ges_problem.u_orbit_opt_ = np.array([[12, 0], [-4, 0], [-4, 0], [-4, 12]])
    ges_problem.x_orbit_opt_ = ges_problem.x_orbit_opt_.transpose()
    ges_problem.u_orbit_opt_ = ges_problem.u_orbit_opt_.transpose()

    # planning horizon
    ges_problem.planning_horizon_n_ = 8
    ges_problem.P_ = 4

    ## robust configuration
    if not(robust == 'none'):
        ges_problem = robust_configurations.get_robust_configuration(ges_problem, robust)


    # return nominal simply supply chain model
    return ges_problem

