# ---------------------------------------------------------------------
# Copyright (c) 2017 Kim Peter Wabersich, email: wabersich@kimpeter.de
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
# -----------------------------------------------------------------
import numpy as np
import simple_supply_chain_model
from casadi import *

## setting '0':
#       In this setting we fix the graph trajectory and only
#       perform the mpc scheme on the resulting periodic LTV system.
#       We assume that we start the the supplier in the graph.

# define controller for constructing the tube
def function_phi(v_real, x_real, z_real, v_graph, x_graph, z_graph):
    # calculate error (explicitely for debugging)
    e_supplier_production_1 = x_real[0] - z_real[0]
    e_supplier_storage_1 = x_real[1] - z_real[1]
    e_truck_storage = x_real[2] - z_real[2]
    e_retailer_storage_1 = x_real[3] - z_real[3]
    e_retailer_storage_2 = x_real[4] - z_real[4]
    e_retailer_storage_3 = x_real[5] - z_real[5]

    # distinguish based on the graph configuration
    if x_graph == 0:  # located at supplier
        u_real = np.array([
                           v_real[0] - e_truck_storage,  # nominal truck loading + truck difference
                           v_real[1]                     # nominal supplier production
                           ])
    if x_graph == 1:  # at retailer 1
        u_real = np.array([
                           v_real[0] + e_retailer_storage_1,  # fill up for retailer error
                           v_real[1] - e_retailer_storage_1,  # produce, what was needed for fill up
                          ])
    if x_graph == 2:  # at retailer 2
        u_real = np.array([
                           v_real[0] + e_retailer_storage_2,  # fill up for retailer error
                           v_real[1] - e_retailer_storage_2,  # produce, what was needed for fill up
                          ])
    if x_graph == 3:  # at retailer 3
        u_real = np.array([
                           v_real[0] + e_retailer_storage_3,  # fill up for retailer error
                           v_real[1] - e_retailer_storage_3,  # produce, what was needed for fill up
                          ])
    return u_real, v_graph

# tightened graph constraints
def function_check_tilde_graph_constraints_tight(z_graph_trajectory, v_graph_trajectory):
    if np.min(v_graph_trajectory) == 1 and np.max(v_graph_trajectory) == 1:
        return True
    else:
        return False


# helper function for creating if statement with logical and operation
def if_kim(cond1,cond2,exp_if,exp_else):
    return if_else(cond1, if_else(cond2,exp_if, exp_else), exp_else)


# helper function for linear integration
# int_a^b (m*x + c) dx
def linear_integration(m,c,a,b):
    return 0.5*m*(b**2 - a**2) + c*(b-a)


## helper function for 2 piece linear integration
# int_a^b l(x) dx with l(x) = m_1*x+c_1 for x<tau and m_2*x+c_2 for x>=tau
def two_piece_linear_integration(m1,c1,m2,c2,tau,a,b):
    return_value = 0
    return_value += if_kim(a>=tau, b>=tau, linear_integration(m2,c2,a,b), 0)
    return_value += if_kim(a<tau, b<tau, linear_integration(m1,c1,a,b), 0)
    return_value += if_kim(a<tau, b>=tau, linear_integration(m1,c1,a,tau)+linear_integration(m2,c2,tau,b), 0)
    return return_value

## helper function for 2 piece linear integration
# int_a^b l(x) dx with l(x) = m_1*x+c_1 for x<tau and m_2*x+c_2 for x>=tau
def two_piece_linear_integration_non_casadi_(m1,c1,m2,c2,tau,a,b):
    return_value = 0
    if a>=tau and b>=tau:
        return_value += linear_integration(m2,c2,a,b)
    if a<tau and b<tau:
        return_value += linear_integration(m1,c1,a,b)
    if a<tau and b>=tau:
        return_value += linear_integration(m1,c1,a,tau)+linear_integration(m2,c2,tau,b)
    return return_value

def function_arr_int_stage_cost_arr_non_casadi(v_real, z_real, v_graph, z_graph, RPI_set_lb, RPI_set_ub):
    # supplier production cost
    stage_cost = linear_integration(1, 0, z_real[0] + RPI_set_lb[0], z_real[0] + RPI_set_ub[0])*6912
    # supplier storage cost
    stage_cost += linear_integration(0.5, 0, z_real[1] + RPI_set_lb[1], z_real[1] + RPI_set_ub[1])*3072
    # truck storage
    stage_cost += linear_integration(1, 0, z_real[2] + RPI_set_lb[2], z_real[2] + RPI_set_ub[2])*2304
    # retailer storage 1
    stage_cost += two_piece_linear_integration_non_casadi_(-10, 0, 1, 0, 0, z_real[3] + RPI_set_lb[3],
                                                           z_real[3] + RPI_set_ub[3])*6912
    # retailer storage 2
    stage_cost += two_piece_linear_integration_non_casadi_(-10, 0, 1, 0, 0, z_real[4] + RPI_set_lb[4],
                                                           z_real[4] + RPI_set_ub[4])*6912
    # retailer storage 3
    stage_cost += two_piece_linear_integration_non_casadi_(-10, 0, 1, 0, 0, z_real[5] + RPI_set_lb[5],
                                                           z_real[5] + RPI_set_ub[5])*6912
    # truck is driving cost
    if v_graph >= 1:
        stage_cost += 1
    return stage_cost/1000.0 # scaling for optimization robustness


# integrated stage cost function
def function_arr_int_stage_cost_arr(v_real, z_real, v_graph, z_graph, RPI_set_lb, RPI_set_ub):
    # supplier production cost
    stage_cost = linear_integration(1, 0, z_real[0]+RPI_set_lb[0], z_real[0]+RPI_set_ub[0])*6912
    # supplier storage cost
    stage_cost += linear_integration(0.5,0,z_real[1]+RPI_set_lb[1], z_real[1]+RPI_set_ub[1])*3072
    # truck storage
    stage_cost += linear_integration(1, 0, z_real[2] + RPI_set_lb[2], z_real[2] + RPI_set_ub[2])*2304
    # retailer storage 1
    stage_cost += two_piece_linear_integration(-10, 0, 1, 0, 0, z_real[3] + RPI_set_lb[3],
                                               z_real[3] + RPI_set_ub[3])*6912
    # retailer storage 2
    stage_cost += two_piece_linear_integration(-10, 0, 1, 0, 0, z_real[4] + RPI_set_lb[4],
                                               z_real[4] + RPI_set_ub[4])*6912
    # retailer storage 3
    stage_cost += two_piece_linear_integration(-10, 0, 1, 0, 0, z_real[5] + RPI_set_lb[5],
                                               z_real[5] + RPI_set_ub[5])*6912
    # truck is driving cost
    stage_cost += if_else(v_graph == 1, 1, 0)
    return np.array([stage_cost/1000.0])  # scaling for optimization robustness


# dummy functions
def function_V_f_dummy(tilde_x_real):
    return SX(0)


def function_X_f_con_dummy(tilde_x_real):
    return SX(0)


# integrated terminal cost function
def function_V_f(x_real, x_graph):
    # we have terminal equality constraints
    return SX(0)


# terminal region
def function_arr_X_f_constraint(x_real, x_graph):
    return SX(0)


# disturbance
def function_disturbance():
    return np.array([
            0,
            0,
            0,
            np.random.randint(-1, 1),
            np.random.randint(-1, 1),
            np.random.randint(-1, 1)])


## definition of robust configuration
def get_robust_configuration(ges_problem, config_ID='0'):

    if config_ID=='0':
        # controller for constructing the tube
        ges_problem.function_phi_ = function_phi
        # tightened state box constraints
        ges_problem.z_lb_tight_ = np.array([0.0,  # supplier items in production >=0
                                            0.0,  # supplier storage >=0
                                            12.0,  # truck storage >=0
                                            -np.inf,
                                            -np.inf,
                                            -np.inf])  # retailer in RR
        ges_problem.z_ub_tight_ = np.array([np.inf,  # supplier items in production >=0
                                            np.inf,  # supplier storage >=0
                                            30,  # truck storage <=30
                                            np.inf,
                                            np.inf,
                                            np.inf])  # retailer in RR
        # tightened input box constraints
        ges_problem.v_lb_tight_ = np.array([-np.inf,  # truck loading bounded by state bounds
                                            0])  # supplier production >=0

        ges_problem.v_ub_tight_ = np.array([np.inf,  # supplier production >=0
                                            np.inf])  # truck loading bounded by state bounds
        # tightened graph constraints
        ges_problem.function_check_tilde_graph_constraints_tight_ \
            = function_check_tilde_graph_constraints_tight
        # RPI set for P-step system
        ges_problem.RPI_set_lb_ = np.array([0.,  # error set supplier production
                                            0.,  # error set supplier storage
                                            -12.,  # error set truck storage
                                            -4.,
                                            -4.,
                                            -4.])  # error set retailer storage
        ges_problem.RPI_set_ub_ = np.array([4.,  # error set supplier production
                                            9.,  # error set supplier storage
                                            0.,  # error set truck storage
                                            0.,
                                            0.,
                                            0.])  # error set retailer storage
        #  Error set W
        ges_problem.W_set_lb = np.array([0,0,0,-1,-1,-1])
        ges_problem.W_set_ub = np.array([0,0,0,0,0,0])
        # integrated stage cost
        ges_problem.num_int_stage_cost_funcs_ = 1
        ges_problem.function_arr_int_stage_cost_arr_ = function_arr_int_stage_cost_arr
        ges_problem.function_arr_int_stage_cost_arr_non_casadi_ = function_arr_int_stage_cost_arr_non_casadi
        # integrated terminal cost function not available
        ges_problem.function_V_f_ = function_V_f
        # terminal region not available
        ges_problem.function_arr_X_f_constraint_ = function_arr_X_f_constraint
        # disturbance
        ges_problem.function_disturbance_ = function_disturbance
        # define optimal periodic orbit
        ges_problem.x_orbit_opt_ = np.array([[12.0000, 0, 12.0000, 3.7727, 4.7727, 5.7727],
                                             [0, 0, 24.0000, 2.7727, 3.7727, 4.7727],
                                             [0.0000, 0, 20.0000, 5.7727, 2.7727, 3.7727],
                                             [0, 0, 16.0000, 4.7727, 5.7727, 2.7727]
                                             ])
        ges_problem.u_orbit_opt_ = np.array([[12, 0], [-4, 0], [-4, 0], [-4, 12]])
        ges_problem.x_orbit_opt_ = ges_problem.x_orbit_opt_.transpose()
        ges_problem.u_orbit_opt_ = ges_problem.u_orbit_opt_.transpose()

    return ges_problem
