# ---------------------------------------------------------------------
# Copyright (c) 2017 Kim Peter Wabersich, email: wabersich@kimpeter.de
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the "Software"),
# to deal in the Software without restriction, including without limitation
# the rights to use, copy, modify, merge, publish, distribute, sublicense,
# and/or sell copies of the Software, and to permit persons to whom the
# Software is furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
# DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
# OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
# DEALINGS IN THE SOFTWARE.
# -----------------------------------------------------------------
import numpy as np
from casadi import *

# simulation of supply chain model
class SimulationSupplyChainModel(object):
    def __init__(self, ges_robust_problem_simulation):
        # number of simulation steps
        self.t_ = 0
        # model
        self.inp_ = ges_robust_problem_simulation
        self.x_real_t_ = ges_robust_problem_simulation.x0_real_
        self.x_graph_t_ = ges_robust_problem_simulation.x0_graph_
        # closed loop information
        self.x_real_history_ = []
        self.u_real_history_ = []
        self.x_graph_history_ = []
        self.u_graph_history_ = []
        # Average performance
        self.avg_performance = 0
        self.avg_robust_performance = 0


    def simulate(self, u_real, u_graph, robust=False):
        # save current state and input
        if self.t_ == 0:
            self.x_real_history_ = np.array([self.x_real_t_])
            self.u_real_history_ = np.array([u_real])
            self.x_graph_history_ = np.array([self.x_graph_t_])
            self.u_graph_history_ = np.array([u_graph])
            self.x_real_history_ = self.x_real_history_.reshape([6, self.t_ + 1])
            self.u_real_history_ = self.u_real_history_.reshape([2, self.t_ + 1])

        # simulate real state variables with normal inputs
        self.x_real_t_ = np.dot(self.inp_.function_matrix_a_dynamics_(SX(self.x_graph_t_)), self.x_real_t_) +\
                         np.dot(self.inp_.function_matrix_b_dynamics_(SX(self.x_graph_t_)), u_real).flatten() + \
                         self.inp_.function_disturbance_nominal_()
        self.x_real_t_ = np.array(self.x_real_t_,dtype=float)
        if robust == True:
            self.x_real_t_ += self.inp_.function_disturbance_()

        # simulate graph state variables
        self.x_graph_t_ = self.inp_.function_graph_dynamics_(self.x_graph_t_, u_graph)


        # save simulation step
        self.x_real_history_ = np.concatenate((self.x_real_history_, np.array([self.x_real_t_]).transpose()), axis=1)
        self.x_graph_history_ = np.concatenate((self.x_graph_history_, np.array([self.x_graph_t_])))
        self.x_real_history_ = self.x_real_history_.reshape([6, self.t_ + 2])
        if self.t_ > 0:
            self.u_real_history_ = np.concatenate((self.u_real_history_, np.array([u_real]).transpose()), axis=1)
            self.u_graph_history_ = np.concatenate((self.u_graph_history_, np.array([u_graph])))
            self.u_real_history_ = self.u_real_history_.reshape([2, self.t_ + 1])

        # count
        self.t_ += 1

    def calculate_avg_cost(self, robust):
        sum_nominal = 0
        sum_int = 0
        for n in range(0, self.t_):
            sum_nominal += max(self.inp_.function_arr_stage_cost_non_casadi_(self.u_real_history_[:,n],
                                                           self.x_real_history_[:,n],
                                                           self.u_graph_history_[n],
                                                           self.x_graph_history_[n]) )
            if robust == True:
                sum_int += self.inp_.function_arr_int_stage_cost_arr_non_casadi_(
                                                                self.u_real_history_[:,n],
                                                               self.x_real_history_[:,n],
                                                               self.u_graph_history_[n],
                                                               self.x_graph_history_[n],
                                                               self.inp_.RPI_set_lb_,
                                                               self.inp_.RPI_set_ub_)
        print("avg nominal cost=" + str(sum_nominal/(self.t_+1)))
        if robust == True:
            print("avg integrated cost=" + str(sum_int /(self.t_+1)))

